"""
DataTransforAPI - classes around API based key lookup.
"""
# pylint: disable=E0401, E0611
import copy
import biothings_client
from biothings.hub.datatransform.datatransform import DataTransform
from biothings.hub.datatransform.datatransform import DataTransformEdge
from biothings.hub.datatransform.datatransform import IDStruct
from biothings.hub.datatransform.datatransform import nested_lookup
from biothings.utils.loggers import get_logger


class DataTransformAPI(DataTransform):
    """
    Perform key lookup or key conversion from one key type to another using
    an API endpoint as a data source.

    This class uses biothings apis to conversion from one key type to another.
    Base classes are used with the decorator syntax shown below::

        @IDLookupMyChemInfo(input_types, output_types)
        def load_document(doc_lst):
            for d in doc_lst:
                yield d

    Lookup fields are configured in the 'lookup_fields' object, examples of which
    can be found in 'IDLookupMyGeneInfo' and 'IDLookupMyChemInfo'.

    Required Options:
        - input_types
            - 'type'
            - ('type', 'nested_source_field')
            - [('type1', 'nested.source_field1'), ('type2', 'nested.source_field2'), ...]
        - output_types:
            - 'type'
            - ['type1', 'type2']

    Additional Options: see DataTransform class
    """
    batch_size = 10
    default_source = '_id'
    lookup_fields = {}

    def __init__(self, input_types, output_types, *args, **kwargs):
        """
        Initialize the IDLookupAPI object.
        """
        self._generate_return_fields()
        super(DataTransformAPI, self).__init__(input_types, output_types, *args, **kwargs)

        # default value of None for client
        self.client = None

        # Keep track of one_to_many relationships
        self.one_to_many_cnt = 0

        self.logger, _ = get_logger('keylookup_api')

    def _valid_input_type(self, input_type):
        """
        Check if the input_type is valid
        :param input_type:
        :return:
        """
        if not isinstance(input_type, str):
            return False
        return input_type.lower() in self.lookup_fields.keys()

    def _valid_output_type(self, output_type):
        """
        Check if the output_type is valid
        :param output_type:
        :return:
        """
        if not isinstance(output_type, str):
            return False
        return output_type.lower() in self.lookup_fields.keys()

    def _generate_return_fields(self):
        """
        Generate the return_fields member variable from the lookup_fields dictionary.
        :return:
        """
        self.return_fields = ''
        for k in self.lookup_fields:
            for field in self._get_lookup_field(k):
                self.return_fields += field + ','
        self.logger.debug("IDLookupAPI return_fields:  {}".format(self.return_fields))

    def key_lookup_batch(self, batchiter):
        """
        Look up all keys for ids given in the batch iterator (1 block)
        :param batchiter:  1 lock of records to look up keys for
        :return:
        """

        id_lst, doc_cache = self._build_cache(batchiter)
        self.logger.info("key_lookup_batch num. id_lst items:  {}".format(len(id_lst)))
        query_res = self._query_many(id_lst)
        qm_struct = self._parse_querymany(query_res)
        return self._replace_keys(qm_struct, doc_cache)

    def _build_cache(self, batchiter):
        """
        Build an id list and document cache for documents read from the
        batch iterator.
        :param batchiter:  an iterator for a batch of documents.
        :return:
        """
        id_lst = []
        doc_cache = []
        for doc in batchiter:

            # handle skip logic
            if self.skip_w_regex and self.skip_w_regex.match(doc['_id']):
                pass
            else:
                for input_type in self.input_types:
                    val = DataTransformAPI._nested_lookup(doc, input_type[1])
                    if val:
                        id_lst.append('"{}"'.format(val))

            # always place the document in the cache
            doc_cache.append(doc)
        return list(set(id_lst)), doc_cache

    def _query_many(self, id_lst):
        """
        Call the biothings_client querymany function with a list of identifiers
        and output fields that will be returned.
        :param id_lst: list of identifiers to query
        :return:
        """
        # Query MyGene.info
        # self.logger.debug("query_many scopes:  {}".format(self.lookup_fields[self.input_type]))
        scopes = []
        for input_type in self.input_types:
            for field in self._get_lookup_field(input_type[0]):
                scopes.append(field)
        client = self._get_client()

        return client.querymany(id_lst,
                                scopes=scopes,
                                fields=self.return_fields,
                                as_generator=True,
                                returnall=True,
                                size=self.batch_size)

    def _parse_querymany(self, query_res):
        """
        Parse the querymany results from the biothings_client into a structure
        that will later be used for document key replacement.
        :param query_res: querymany results
        :return:
        """
        # self.logger.debug("QueryMany Structure:  {}".format(query_res))
        qm_struct = {}
        for q_out in query_res['out']:
            query = q_out['query']
            val = self._parse_h(q_out)
            if val:
                if query not in qm_struct.keys():
                    qm_struct[query] = [val]
                else:
                    self.one_to_many_cnt += 1
                    qm_struct[query] = qm_struct[query] + [val]
        # self.logger.debug("parse_querymany num qm_struct keys: {}"\
        #        .format(len(qm_struct.keys())))
        # self.logger.info("parse_querymany running one_to_many_cnt: {}"\
        #        .format(self.one_to_many_cnt))
        # self.logger.debug("parse_querymany qm_struct: {}"\
        #        .format(qm_struct.keys()))
        return qm_struct

    def _parse_h(self, hit):
        """
        Parse a single hit from the API.
        :param hit:
        :return: dictionary of keys
        """
        for output_type in self.output_types:
            for doc_field in self._get_lookup_field(output_type):
                val = DataTransformAPI._nested_lookup(hit, doc_field)
                if val:
                    return val
        return None

    def _replace_keys(self, qm_struct, doc_cache):
        """
        Build a new list of documents to return that have their keys
        replaced by answers specified in the qm_structure which
        was built earlier.
        :param qm_struct: structure of keys from _parse_querymany
        :param doc_cache: cache of documents that will have keys replaced.
        :return:
        """
        # Replace the keys and build up a new result list
        res_lst = []
        for doc in doc_cache:
            new_doc = None
            for input_type in self.input_types:
                # doc[input_type[1]] must be typed to a string because
                # qm_struct.keys are always strings
                val = DataTransformAPI._nested_lookup(doc, input_type[1])
                if val in qm_struct.keys():
                    for key in qm_struct[val]:
                        new_doc = copy.deepcopy(doc)
                        new_doc['_id'] = key
                        res_lst.append(new_doc)
                # Break out if an input type was used.
                if new_doc:
                    break
            if not new_doc and (
                    (self.skip_w_regex and self.skip_w_regex.match(doc['_id']))
                    or not self.skip_on_failure):
                res_lst.append(doc)

        self.logger.info("_replace_keys:  Num of documents yielded:  {}".format(len(res_lst)))
        # Yield the results
        for res in res_lst:
            yield res

    def _get_lookup_field(self, field):
        """
        Getter for lookup fields which may be either a string or a list of string fields.
        :param field: the name of the field to lookup
        :return:
        """
        if field not in self.lookup_fields.keys():
            raise KeyError('provided field {} is not in self.lookup_fields'.format(field))
        if isinstance(self.lookup_fields[field], str):
            return [self.lookup_fields[field]]
        return self.lookup_fields[field]

    def _get_client(self):
        """get biothings_client"""
        raise NotImplementedError("_get_client not implemented in the super class")


class DataTransformMyChemInfo(DataTransformAPI):
    """Single key lookup for MyChemInfo"""
    lookup_fields = {
        'unii': 'unii.unii',
        'rxnorm': [
            'unii.rxcui'
        ],
        'drugbank': 'drugbank.drugbank_id',
        'chebi': 'chebi.chebi_id',
        'chembl': 'chembl.molecule_chembl_id',
        'pubchem': 'pubchem.cid',
        'drugname': [
            'drugbank.name',
            'unii.preferred_term',
            'chebi.chebi_name',
            'chembl.pref_name'
        ],
        'inchi': [
            'drugbank.inchi',
            'chembl.inchi',
            'pubchem.inchi'
        ],
        'inchikey': [
            'drugbank.inchi_key',
            'chembl.inchi_key',
            'pubchem.inchi_key'
        ]
    }
    # The order of output_types decides the priority
    # of the key types we used to get _id value
    output_types = ['inchikey', 'unii', 'rxnorm', 'drugbank',
                    'chebi', 'chembl', 'pubchem', 'drugname'],

    def __init__(self, input_types,
                 output_types=None,
                 skip_on_failure=False,
                 skip_w_regex=None):
        """
        Initialize the class by seting up the client object.
        """
        _output_types = output_types or self.output_types
        super(DataTransformMyChemInfo, self).__init__(
            input_types, _output_types, skip_on_failure, skip_w_regex)

    def _get_client(self):
        """
        Get Client - return a client appropriate for IDLookup

        This method must be defined in the child class.  It is an artifact
        of multithreading.
        :return:
        """
        if not self.client:
            self.client = biothings_client.get_client('drug')
        return self.client


class BiothingsAPIEdge(DataTransformEdge):
    """
    APIEdge - IDLookupEdge object for API calls
    """
    # define in subclass
    client_name = None

    def __init__(self, lookup, fields, weight=1, label=None, url=None):
        # pylint: disable=R0913
        super(BiothingsAPIEdge, self).__init__(label)
        self.init_state()
        if isinstance(lookup, str):
            self.scopes = [lookup]
        elif isinstance(lookup, list):
            self.scopes = lookup
        else:
            raise TypeError("scopes argument must be str or list")
        if isinstance(fields, str):
            self.fields = [fields]
        elif isinstance(fields, list):
            self.fields = fields
        else:
            raise TypeError("fields argument must be str or list")
        self.weight = weight
        self.url = url

    def init_state(self):
        """initialize state - pickleable member variables"""
        self._state = {
            "client": None,
            "logger": None
        }

    @property
    def client(self):
        """property getter for client"""
        if not self._state["client"]:
            try:
                self.prepare_client()
            except NotImplementedError:
                # if accessed but not ready, then just ignore and return invalid value for a client
                return None
        return self._state["client"]

    def prepare_client(self):
        """
        Load the biothings_client for the class
        :return:
        """
        if not self.client_name:
            raise NotImplementedError("Define client_name in subclass")
        if self.url:
            self._state["client"] = biothings_client.get_client(self.client_name, url=self.url)
        else:
            self._state["client"] = biothings_client.get_client(self.client_name)
        self.logger.info("Registering biothings_client {}".format(self.client_name))

    def edge_lookup(self, keylookup_obj, id_strct, debug=False):
        """
        Follow an edge given a key.

        This method uses the data in the edge_object
        to find one key to another key using an api.
        :param edge:
        :param key:
        :return:
        """
        # If no keys were passed return an empty idstruct_class
        # pylint: disable=C1801
        if not len(id_strct):
            return keylookup_obj.idstruct_class()
        # query the api
        query_res = self._query_many(keylookup_obj, id_strct)
        new_id_strct = self._parse_querymany(keylookup_obj, query_res, id_strct, self.fields, debug)
        return new_id_strct

    def _query_many(self, keylookup_obj, id_strct):
        """
        Call the biothings_client querymany function with a list of identifiers
        and output fields that will be returned.
        :param id_lst: list of identifiers to query
        :return:
        """
        if not isinstance(id_strct, IDStruct):
            raise TypeError("id_strct shouldb be of type IDStruct")
        id_lst = []
        for key in id_strct.id_lst:
            id_lst.append('"{}"'.format(key))
        return self.client.querymany(id_lst,
                                     scopes=self.scopes,
                                     fields=self.fields,
                                     as_generator=True,
                                     returnall=True,
                                     size=keylookup_obj.batch_size,
                                     verbose=False)

    def _parse_querymany(self, keylookup_obj, query_res, id_strct, fields, debug):
        # pylint: disable=R0913, W0613
        """
        Parse the querymany results from the biothings_client into a structure
        that will later be used for document key replacement.
        :param query_res: querymany results
        :return:
        """
        # self.logger.debug("QueryMany Structure:  {}".format(query_res))
        qm_struct = IDStruct()

        # Keep the old debug information
        if debug:
            qm_struct.import_debug(id_strct)

        # pylint: disable=R1702
        for q_out in query_res['out']:
            query = q_out['query']
            for field in fields:
                val = nested_lookup(q_out, field)
                if val:
                    for (orig_id, curr_id) in id_strct:
                        # query is always a string, so this check requires conversion
                        if query == str(curr_id):
                            qm_struct.add(orig_id, val)
                            # save debug information in the option is set
                            if debug:
                                qm_struct.set_debug(orig_id, self.label, val)
        return qm_struct


class MyChemInfoEdge(BiothingsAPIEdge):
    """
    The MyChemInfoEdge uses the MyChem.info API to convert identifiers.
    """
    client_name = "drug"

    def __init__(self, lookup, field, weight=1, label=None, url=None):
        # pylint: disable=R0913, W0235
        """
        :param lookup: The field in the API to search with the input identifier.
        :type lookup: str
        :param field: The field in the API to convert to.
        :type field: str
        :param weight: Weights are used to prefer one path over another.
                       The path with the lowest weight is preferred.
                       The default weight is 1.
        :type weight: int
        """
        super(MyChemInfoEdge, self).__init__(lookup, field, weight, label, url)


class MyGeneInfoEdge(BiothingsAPIEdge):
    """
    The MyGeneInfoEdge uses the MyGene.info API to convert identifiers.
    """
    client_name = "gene"

    def __init__(self, lookup, field, weight=1, label=None, url=None):
        # pylint: disable=R0913, W0235
        """
        :param lookup: The field in the API to search with the input identifier.
        :type lookup: str
        :param field: The field in the API to convert to.
        :type field: str
        :param weight: Weights are used to prefer one path over another.
                       The path with the lowest weight is preferred.
                       The default weight is 1.
        :type weight: int
        """
        super(MyGeneInfoEdge, self).__init__(lookup, field, weight, label, url)


####################

class DataTransformMyGeneInfo(DataTransformAPI):
    """deprecated"""
    lookup_fields = {
        'ensembl': 'ensembl.gene',
        'entrezgene': 'entrezgene',
        'symbol': 'symbol',
        'uniprot': 'uniprot.Swiss-Prot'
    }

    def __init__(self, input_types,
                 output_types=['entrezgene'],
                 skip_on_failure=False,
                 skip_w_regex=None):
        # pylint: disable=W0102
        """
        Initialize the class by seting up the client object.
        """
        super(DataTransformMyGeneInfo, self).__init__(
            input_types, output_types, skip_on_failure, skip_w_regex)

    def _get_client(self):
        """
        Get Client - return a client appropriate for IDLookup

        This method must be defined in the child class.  It is an artifact
        of multithreading.
        :return:
        """
        if not self.client:
            self.client = biothings_client.get_client('gene')
        return self.client



'''
Deprecated. This module is not used any more.
'''

from biothings.utils.mongo import get_src_db
from biothings.utils.common import loadobj


class MongoSync():
    def __init__(self):
        self._src = get_src_db()

    def add_update(self, source, merge_collection, ids):
        for _id in ids:
            doc = self._src[source].get_from_id(_id)
            self._src[merge_collection].update({'_id': _id}, {'$set': doc}, upsert=True)

    def delete(self, merge_collection, field, ids):
        for _id in ids:
            doc = self._src[merge_collection].get_from_id(_id)
            if set(doc) == set(['_id', field]) or set(doc) == set(['_id', field, 'snpeff', 'vcf']):
                self._src[merge_collection].remove(_id)
            else:
                self._src[merge_collection].update({'_id': _id}, {'$unset': {field: 1}})

    def main(self, diff_filepath, merge_collection, field):
        diff = loadobj(diff_filepath)
        source_collection = diff['source']
        add_ids = diff['add']
        delete_ids = diff['delete']
        update_ids = [_doc['_id'] for _doc in diff['update']]
        self.add_update(source_collection, merge_collection, add_ids)
        self.add_update(source_collection, merge_collection, update_ids)
        self.delete(merge_collection, field, delete_ids)



'''
    Utils for creating swagger spec from a config_web module.

'''
from collections import OrderedDict
import yaml

def setup_yaml():
    """ https://stackoverflow.com/a/8661021 """
    represent_dict_order = lambda self, data:  self.represent_mapping('tag:yaml.org,2002:map', data.items())
    yaml.add_representer(OrderedDict, represent_dict_order)

all_params = {}

def _get_common_parameters(settings, biothing_object, url):
    _all = _get_all_params(settings)
    _ret = OrderedDict([])
    # get all params except for q, scopes, ids -> these are added to the request body, and not the query string
    _common_params_list = [(k,v) for (k,v) in _all.items() if (k != 'q') and (k != 'scopes') and (k != 'ids') and (k in getattr(settings, 'KWARG_DESCRIPTIONS', {}))]
    # add 'special' parameters not controlled by main biothings web pipeline
    _common_params_list.extend([('callback', {'type': str, 'default': None}),
                                ('email', {'type': str, 'default': None})])
    _description_templates = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    for (param, description) in _common_params_list:
        _description_template_object = _description_templates.get(param)
        _param_display = _description_template_object.get('name')
        _ret.update(OrderedDict([
            (_param_display, OrderedDict([
                ('name', _param_display),
                ('in', 'query'),
                ('description', _description_template_object.get('text_template', '').format(
                    biothing_object=biothing_object, 
                    param_type=_string_types(_all.get(param, {}).get('type', '')),
                    param_default_value=_param_default_values(_all.get(param, {}).get('default', '')), 
                    param_max=_param_max(_all.get(param, {}).get('max', '')))),
                ('schema', {'type': _schema_types(_all.get(param, {}).get('type', ''))})
                ]))
            ]))
    return _ret
              
def _get_schemas(settings, biothing_object, url):
    return OrderedDict([
        ("string_or_array", {
            "oneOf": [
                {"type": "array", "items": {"type": "string"}},
                {"type": "string"}
            ]
        }),
        ("int64_or_array", {
            "oneOf": [
                {"type": "array", "items": {"type": "integer", "format": "int64"}},
                {"type": "integer", "format": "int64"}
            ]
        }),
        ("QueryResult", {
            "type": "object",
            "properties": OrderedDict([
                ("max_score", {"type": "number", "format": "float"}),
                ("took", {"type": "integer"}),
                ("total", {"type": "integer"}),
                ("hits", {"type": "array", "items": {"$ref": "#/components/schemas/{}".format(biothing_object.title())}})
            ])
        }),
        ("QueryPOSTResult", {
            "type": "array",
            "items": {"allOf": [
                {"$ref": '#/components/schemas/{}'.format(biothing_object.title())},
                {"type": "object", "properties": {"_score": {"type": "number", "format": "float"}, "query": {"type": "string"}}}
            ]}}),
        ("ErrorResult", {"type": "object",
            "properties": {"success": {"type": "boolean"}, "message": {"type": "string"}}}),
        ("{}".format(biothing_object.title()), {"type": "object", "required": ["_id"], "properties": {"_id":{"type": "string"}}})
        ])

def _schema_types(t):
    if t == type(bool()):
        return 'boolean'
    if t == type(int()):
        return 'integer'
    return 'string'

def _string_types(t):
    if t == type(str()):
        return ' Type: string.'
    elif t == type(int()):
        return ' Type: integer.'
    elif t == type([]):
        return ' Type: string (list).'
    elif t == type(bool()):
        return ' Type: boolean.'
    return ' Type: string.'

def _param_default_values(d):
    if not d:
        return ' Default: None.'
    if type(d) == type(int()) or type(d) == type(str()):
        return ' Default: {}.'.format(d)
    elif type(d) == type(True) and d == True:
        return ' Default: true.'
    elif type(d) == type(False) and d == False:
        return ' Default: false.'
    return ' Default: None.'

def _param_max(m):
    if not m:
        return ''
    return ' Max: {}.'.format(m)

def _query_get_params(settings):
    _ret = {}
    for d in [settings.QUERY_GET_CONTROL_KWARGS, settings.QUERY_GET_ES_KWARGS,
        settings.QUERY_GET_ESQB_KWARGS, settings.QUERY_GET_TRANSFORM_KWARGS]:
        for k, v in d.items():
            if k != 'q' and k in getattr(settings, 'KWARG_DESCRIPTIONS', {}):
                _ret.setdefault(k, v)

    _ret['callback'] = {'type': str, 'default': None}
    _ret['email'] = {'type': str, 'default': None}

    return _ret

def _metadata_params(settings):
    _ret = {}
    for d in [settings.METADATA_GET_CONTROL_KWARGS, settings.METADATA_GET_ES_KWARGS,
        settings.METADATA_GET_ESQB_KWARGS, settings.METADATA_GET_TRANSFORM_KWARGS]:
        for k, v in d.items():
            if k in getattr(settings, 'KWARG_DESCRIPTIONS', {}):
                _ret.setdefault(k, v)

    _ret['callback'] = {'type': str, 'default': None}

    return _ret

def _annotation_get_params(settings):
    _ret = {}
    for d in [settings.ANNOTATION_GET_CONTROL_KWARGS, settings.ANNOTATION_GET_ES_KWARGS,
        settings.ANNOTATION_GET_ESQB_KWARGS, settings.ANNOTATION_GET_TRANSFORM_KWARGS]:
        for k, v in d.items():
            if k != 'q' and k in getattr(settings, 'KWARG_DESCRIPTIONS', {}):
                _ret.setdefault(k, v)

    _ret['callback'] = {'type': str, 'default': None}
    _ret['email'] = {'type': str, 'default': None}

    return _ret

def _query_post_params(settings):
    _ret = {}
    for d in [settings.QUERY_POST_CONTROL_KWARGS, settings.QUERY_POST_ES_KWARGS,
        settings.QUERY_POST_ESQB_KWARGS, settings.QUERY_POST_TRANSFORM_KWARGS]:
        for k, v in d.items():
            if k != 'q' and k != 'scopes' and k in getattr(settings, 'KWARG_DESCRIPTIONS', {}):
                _ret.setdefault(k, v)

    _ret['email'] = {'type': str, 'default': None}

    return _ret

def _annotation_post_params(settings):
    _ret = {}
    for d in [settings.ANNOTATION_POST_CONTROL_KWARGS, settings.ANNOTATION_POST_ES_KWARGS,
        settings.ANNOTATION_POST_ESQB_KWARGS, settings.ANNOTATION_POST_TRANSFORM_KWARGS]:
        for k, v in d.items():
            if k != 'ids' and k in getattr(settings, 'KWARG_DESCRIPTIONS', {}):
                _ret.setdefault(k, v)

    _ret['email'] = {'type': str, 'default': None}

    return _ret

def _get_all_params(settings):
    # returns all possible parameters for a biothing web settings object
    if all_params:
        return all_params

    for d in [settings.ANNOTATION_GET_CONTROL_KWARGS, settings.ANNOTATION_GET_ES_KWARGS,
        settings.ANNOTATION_GET_ESQB_KWARGS, settings.ANNOTATION_GET_TRANSFORM_KWARGS,
        settings.ANNOTATION_POST_CONTROL_KWARGS, settings.ANNOTATION_POST_ES_KWARGS,
        settings.ANNOTATION_POST_ESQB_KWARGS, settings.ANNOTATION_POST_TRANSFORM_KWARGS,
        settings.QUERY_GET_CONTROL_KWARGS, settings.QUERY_GET_ES_KWARGS,
        settings.QUERY_GET_ESQB_KWARGS, settings.QUERY_GET_TRANSFORM_KWARGS,
        settings.QUERY_POST_CONTROL_KWARGS, settings.QUERY_POST_ES_KWARGS,
        settings.QUERY_POST_ESQB_KWARGS, settings.QUERY_POST_TRANSFORM_KWARGS,
        settings.METADATA_GET_CONTROL_KWARGS, settings.METADATA_GET_ES_KWARGS,
        settings.METADATA_GET_ESQB_KWARGS, settings.METADATA_GET_TRANSFORM_KWARGS]:
        for k, v in d.items():
            all_params.setdefault(k, v)

    return all_params

def _query_get_description(settings, biothing_object, url):
    _descriptions = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    _parameters = [OrderedDict([
        ('name', 'q'),
        ('in', 'query'),
        ('description', _descriptions.get('q',{}).get('text_template').format(doc_query_syntax_url='')),
        ('required', True),
        ('example', getattr(settings, 'STATUS_CHECK', {}).get('id', 'EXAMPLE')),
        ('schema', {'type': 'string'})
    ])]
    for k in _query_get_params(settings).keys():
        _parameters.append(
            OrderedDict([
                ('name', _descriptions.get(k).get('name')),
                ('$ref', '#/components/parameters/{}'.format(_descriptions.get(k).get('name')))
            ]))
    return OrderedDict([
        ('tags', ["query"]),
        ('summary', 'Make {biothing_object} queries and return matching {biothing_object} hits. Supports JSONP and CORS as well.'.format(biothing_object=biothing_object)),
        ('parameters', _parameters),
        ('responses', OrderedDict([
            ('200', OrderedDict([
                ('description', 'A query response with the "hits" property'),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/QueryResult'}}})
            ])),
            ('400', OrderedDict([
                ('description', 'A response indicating an improperly formatted query'),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/ErrorResult'}}})
            ]))    
        ]))])

def _query_post_description(settings, biothing_object, url):
    _descriptions = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    _parameters = []
    for k in _query_post_params(settings).keys():
        _parameters.append(
            OrderedDict([
                ('name', _descriptions.get(k).get('name')),
                ('$ref', '#/components/parameters/{}'.format(_descriptions.get(k).get('name')))
            ]))
    return OrderedDict([
        ('tags', ["query"]),
        ('summary', 'Make {biothing_object} batch queries and return matching {biothing_object} hits'.format(biothing_object=biothing_object)),
        ('requestBody', {'content': {'application/x-www-form-urlencoded': {'schema': {'properties': OrderedDict([
            ('q', OrderedDict([
                ('description', 'multiple query terms separated by comma (also "+" or whitespace).  Does not support wildcard queries'),
                ('type', 'string')
            ])),
            ('scopes', OrderedDict([
                ('description', _descriptions.get('scopes', {}).get('text_template', '').format(
                    param_type='string', param_max='')),
                ('type', 'string')
            ]))
        ])}}}}),
        ('parameters', _parameters),
        ('responses', OrderedDict([
            ('200', OrderedDict([
                ('description', 'Query response objects with the "hits" property'),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/QueryPOSTResult'}}})
            ])),
            ('400', OrderedDict([
                ('description', 'A response indicating an improperly formatted query'),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/ErrorResult'}}})
            ]))]))    
    ])

def _annotation_get_description(settings, biothing_object, url):
    _descriptions = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    _parameters = [
        OrderedDict([
            ('name', '{biothing_object}id'.format(biothing_object=biothing_object)),
            ('in', 'path'),
            ('description', '{biothing_object} id'.format(biothing_object=biothing_object)),
            ('required', True),
            ('example', getattr(settings, 'STATUS_CHECK', {}).get('id', 'EXAMPLE')),
            ('schema', {'type': 'string'})
        ])
    ]
    for k in _annotation_get_params(settings).keys():
        _parameters.append(
            OrderedDict([
                ('name', _descriptions.get(k).get('name')),
                ('$ref', '#/components/parameters/{}'.format(_descriptions.get(k).get('name')))
            ]))
    return OrderedDict([
        ('tags', ["{}".format(biothing_object)]),
        ('summary', 'Retrieve {biothing_object} objects based on {biothing_object} id.  Supports JSONP and CORS as well.'.format(biothing_object=biothing_object)),
        ('parameters', _parameters),
        ('responses', OrderedDict([
            ('200', OrderedDict([
                ('description', 'A matching {biothing_object} object'.format(biothing_object=biothing_object)),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/{}'.format(biothing_object.title())}}})
            ])),
            ('404', OrderedDict([
                ('description', 'A response indicating an unknown {biothing_object} id'.format(biothing_object=biothing_object))
            ]))]))    
    ])

def _annotation_post_description(settings, biothing_object, url):
    _descriptions = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    _parameters = []
    for k in _annotation_post_params(settings).keys():
        _parameters.append(
            OrderedDict([
                ('name', _descriptions.get(k).get('name')),
                ('$ref', '#/components/parameters/{}'.format(_descriptions.get(k).get('name')))
            ]))
    return OrderedDict([
        ('tags', ["{biothing_object}".format(biothing_object=biothing_object)]),
        ('summary', 'For a list of {biothing_object} ids, return the matching {biothing_object} object'.format(biothing_object=biothing_object)),
        ('requestBody', {'content': {'application/x-www-form-urlencoded': {'schema': OrderedDict([
            ('properties', {'ids': OrderedDict([
                ('description', _descriptions.get('ids', {}).get('text_template', '').format(
                    biothing_object=biothing_object, param_type=' Type: string (list).', param_default_value='',
                    param_max = _param_max(getattr(settings, 'ANNOTATION_POST_CONTROL_KWARGS', {}).get('max', 1000)))),
                ('type', 'string')
            ])}),
            ('required', ['ids'])])}}}),
        ('parameters', _parameters),
        ('responses', OrderedDict([
            ('200', OrderedDict([
                ('description', 'A list of matching {biothing_object} objects'.format(biothing_object=biothing_object)),
                ('content', {'application/json': {'schema': {'type': 'array', 'items': {'$ref': '#/components/schemas/{}'.format(biothing_object.title())}}}})
            ])),
            ('400', OrderedDict([
                ('description', 'A response indicating an improperly formatted query'),
                ('content', {'application/json': {'schema': {'$ref': '#/components/schemas/ErrorResult'}}})
            ]))]))    
    ])

def _metadata_description(settings, biothing_object, url):
    return OrderedDict([
        ('tags', ['metadata']),
        ('summary', 'Get metadata about the data available from {url}'.format(url=url)),
        ('parameters', [OrderedDict([('name', 'callback'), ('in', 'query'), ('$ref', '#/components/parameters/callback')])]),
        ('responses', {'200': {'description': '{url} metadata object'.format(url=url)}})
    ])

def _metadata_fields_description(settings, biothing_object, url):
    _descriptions = getattr(settings, 'KWARG_DESCRIPTIONS', {})
    _parameters = []
    for k in _metadata_params(settings).keys():
        _parameters.append(
            OrderedDict([
                ('name', _descriptions.get(k).get('name')),
                ('$ref', '#/components/parameters/{}'.format(_descriptions.get(k).get('name')))
            ]))
    return OrderedDict([
        ('tags', ['metadata']),
        ('summary', 'Get metadata about the data fields available from a {url} {biothing_object} object'.format(biothing_object=biothing_object, url=url)),
        ('parameters', _parameters),
        ('responses', {'200': {'description': '{url} metadata fields object'.format(url=url)}})
    ])

def create_openapi_json(_settings, **kwargs):
    _url = getattr(_settings, 'GA_TRACKER_URL', 'MyBiothing.info')
    _biothing_object = getattr(_settings, 'ES_DOC_TYPE', 'biothing')
    _doc = OrderedDict([
        ("openapi", kwargs.get('openapi', '3.0.0')),
        ("info", OrderedDict([
            ("version", kwargs.get('info__version', '{:.1f}'.format(float(getattr(_settings, 'API_VERSION', 'v1').lstrip('v'))))),
            ("title", kwargs.get('info__title', '{} API'.format(getattr(_settings,'GA_TRACKER_URL', 'MyBiothing.info')))),
            ("description", kwargs.get('info__description', getattr(_settings, 'KWARG_DESCRIPTION', {}).get('_root', 'Documentation of the {url} {biothing_object} query web services.  Learn more about [{url}](http://{url}/)'.format(url=_url, biothing_object=_biothing_object)))),
            ("termsOfService", kwargs.get('info__termsOfService', 'http://{url}/terms'.format(url=_url))),
            ("contact", OrderedDict([
                ("name", kwargs.get('info__contact__name', 'Chunlei Wu')),
                ("x-role", kwargs.get('info__contact__x_role', 'responsible developer')),
                ("email", kwargs.get('info__contact__email', 'help@biothings.io')),
                ("x-id", kwargs.get('info__contact__x_id', 'https://github.com/newgene'))
            ]))
        ])),
        ("servers", [OrderedDict([
            ("url", "http://{url}/{version}".format(url=_url, version=getattr(_settings, 'API_VERSION', 'v1'))),
            ("description", "Production server")
        ]),
        OrderedDict([
            ("url", "https://{url}/{version}".format(url=_url, version=getattr(_settings, 'API_VERSION', 'v1'))),
            ("description", "Encrypted Production server")
        ])]),
        ("tags", [{"name": "{biothing_object}".format(biothing_object=_biothing_object)},
                  {"name": "query"},
                  {"name": "metadata"}]),
        ("paths", OrderedDict([
            ("/{biothing_object}/{{{biothing_object}id}}".format(biothing_object=_biothing_object), 
            # annotation GET
            OrderedDict([
                ("get", _annotation_get_description(settings=_settings, biothing_object=_biothing_object, url=_url))
            ])),
            # annotation POST
            ("/{biothing_object}".format(biothing_object=_biothing_object), OrderedDict([
                ("post", _annotation_post_description(settings=_settings, biothing_object=_biothing_object, url=_url))
            ])),
            # query GET/POST
            ("/query", OrderedDict([
                ("get", _query_get_description(settings=_settings, biothing_object=_biothing_object, url=_url)),
                ("post", _query_post_description(settings=_settings, biothing_object=_biothing_object, url=_url))
            ])),
            # metadata GET
            ("/metadata", OrderedDict([
                ("get", _metadata_description(settings=_settings, biothing_object=_biothing_object, url=_url))
            ])),
            # metadata/fields GET
            ("/metadata/fields", OrderedDict([
                ("get", _metadata_fields_description(settings=_settings, biothing_object=_biothing_object, url=_url))
            ]))
        ])),
        ("components", OrderedDict([
            ("parameters", _get_common_parameters(settings=_settings, biothing_object=_biothing_object, url=_url)),
            ("schemas", _get_schemas(settings=_settings, biothing_object=_biothing_object, url=_url))
        ]))
    ])
    return _doc

def create_openapi_yaml(settings, yaml_file, **kwargs):
    ''' Create an openapi compliant spec for the biothing project whose settings are input.

        settings - a biothings web settings object
        yaml_file - the output file
    '''
    setup_yaml()
    _json = create_openapi_json(settings, **kwargs)
    with open(yaml_file, 'w') as outf:
        yaml.dump(_json, outf, default_flow_style=False)



import importlib
import threading
import re
import copy
import asyncio
import os
import types
import glob
import time
import datetime
from functools import wraps, partial
from pprint import pformat
from collections import OrderedDict
import concurrent.futures

import aiocron
import dill as pickle
import psutil

from biothings import config
logger = config.logger

from biothings.utils.mongo import get_src_conn
from biothings.utils.common import timesofar, get_random_string, sizeof_fmt


def track(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        job_id = args[0]
        ptype = args[1]  # tracking process or thread ?
        # we're looking for some "pinfo" value (process info) to later
        # reporting. If we can't find any, we'll try our best to figure out
        # what this is about...
        # func is the do_work wrapper, we want the actual partial
        # is first arg a callable (func) or pinfo ?
        if callable(args[2]):
            innerfunc = args[2]
            innerargs = args[3:]
            pinfo = None
        else:
            innerfunc = args[3]
            innerargs = args[4:]
            # we want to let the original as-is, as it can still contain
            # usefull information like predicates
            pinfo = copy.deepcopy(args[2])

        # predicates can't be pickles/dilled
        pinfo.pop("__predicates__", None)
        # just informative, so stringify is just ok there)
        # make sure we can pickle the whole thing (and it's
        innerargs = [str(arg) for arg in innerargs]
        if type(innerfunc) == partial:
            fname = innerfunc.func.__name__
        elif type(innerfunc) == types.MethodType:
            fname = innerfunc.__self__.__class__.__name__
        else:
            fname = innerfunc.__name__

        firstarg = innerargs and innerargs[0] or ""
        if not pinfo:
            pinfo = {"category": None,
                     "source": None,
                     "step": None,
                     "description": "%s %s" % (fname, firstarg)}

        pinfo["started_at"] = time.time()
        worker = {
            'func_name': fname,
            'args': innerargs,
            'kwargs': kwargs,
            'job': pinfo
        }
        results = None
        exc = None
        trace = None
        pidfile = None
        try:
            _id = None
            if ptype == "thread":
                _id = "%s" % threading.current_thread().getName()
            else:
                _id = os.getpid()
            # add random chars: 2 jobs handled by the same slot (pid or thread)
            # would override filename otherwise
            fn = "%s_%s" % (_id, job_id)
            worker["job"]["id"] = _id
            pidfile = os.path.join(config.RUN_DIR, "%s.pickle" % fn)
            pickle.dump(worker, open(pidfile, "wb"))
            results = func(*args, **kwargs)
        except Exception as e:
            import traceback
            trace = traceback.format_exc()
            logger.error("err %s\n%s", e, trace)
            # we want to store exception so for now, just make a reference
            exc = e
        finally:
            if pidfile and os.path.exists(pidfile):
                logger.debug("Remove PID file '%s'", pidfile)
                os.unlink(pidfile)
        # now raise original exception
        if exc:
            raise exc
        return results
    return func_wrapper

@track
def do_work(job_id, ptype, pinfo=None, func=None, *args, **kwargs):
    # pinfo is optional, and func is not. and args and kwargs must
    # be after func. just to say func is mandatory, despite what the
    # signature says
    assert func
    # need to wrap calls otherwise multiprocessing could have
    # issue pickling directly the passed func because of some import
    # issues ("can't pickle ... object is not the same as ...")
    return func(*args, **kwargs)


def find_process(pid):
    g = psutil.process_iter()
    for p in g:
        if p.pid == pid:
            break
    return p


class UnknownResource(Exception):
    pass
class ResourceError(Exception):
    pass
class ManagerError(Exception):
    pass
class ResourceNotFound(Exception):
    pass

class BaseManager(object):

    def __init__(self, job_manager):
        self.register = {}
        self.job_manager = job_manager
        self.clean_stale_status()

    def clean_stale_status(self):
        """
        During startup, search for action in progress which would have
        been interrupted and change the state to "canceled".
        Ex: some donwloading processes could have been interrupted, at
        startup, "downloading" status should be changed to "canceled" so
        to reflect actual state on these datasources.
        This must be overriden in subclass.
        """

    def __repr__(self):
        registered = sorted(list(self.register.keys()))
        return "<%s [%d registered]: %s>" % (self.__class__.__name__, len(self.register), registered)

    def __getitem__(self, src_name):
        try:
            # as a main-source
            return self.register[src_name]
        except KeyError:
            try:
                # as a sub-source
                main, sub = src_name.split(".")
                srcs = self.register[main]
                # there can be many uploader for one resource (when each is dealing
                # with one specific file but upload to the same collection for instance)
                # so we want to make sure user is aware of this and not just return one
                # uploader when many are needed
                # on the other hand, if only one avail, just return it
                res = []
                for src in srcs:
                    if src.name == sub:
                        res.append(src)
                if len(res) == 0:
                    raise KeyError(src_name)
                else:
                    return res
            except (ValueError, AttributeError, KeyError):
                # nope, can't find it...
                raise KeyError(src_name)

    def poll(self, state, func, col):
        '''
        Search for source in collection 'col' with a pending flag list
        containing 'state' and and call 'func' for each document found
        (with doc as only param)
        '''
        if not self.poll_schedule:
            raise ManagerError("poll_schedule is not defined")

        @asyncio.coroutine
        def check_pending(state):
            # sources = [src for src in col.find({'pending': state}) if type(src['_id']) == str]   # TODO: remove this line
            sources = [src for src in col.find({'pending': state}) if isinstance(src['_id'], str)]
            if sources:
                logger.info(
                    "Found %d resources with pending flag %s (%s)",
                    len(sources), state, repr([src["_id"] for src in sources])
                )
            for src in sources:
                logger.info("Run %s for pending flag %s on source '%s'", func, state, src["_id"])
                try:
                    # first reset flag to make sure we won't call func multiple time
                    col.update({"_id": src["_id"]}, {"$pull": {"pending": state}})
                    func(src)
                except ResourceNotFound:
                    logger.error("Resource '%s' has a pending flag set to %s but is not registered in manager",
                                 src["_id"], state)
        cron = aiocron.crontab(
            self.poll_schedule, func=partial(check_pending, state),
            start=True, loop=self.job_manager.loop
        )


class BaseStatusRegisterer(object):

    def load_doc(self, key_name, stage):
        """
        Find document using key_name and stage, stage being a
        key within the document matching a specific process name:
        Ex: {"_id":"123","snapshot":"abc"}
            load_doc("abc","snapshot")
        will return the document. Note key_name is first used to
        find the doc by its _id.
        Ex: with another doc {"_id" : "abc", "snapshot" : "somethingelse"}
            load_doc{"abc","snapshot")
        will return doc with _id="abc", not "123"
        """
        doc = self.collection.find_one({'_id': key_name})
        if not doc:
            doc = []
            bdocs = self.collection.find()
            for adoc in bdocs:
                if key_name in adoc.get(stage, {}):
                    doc.append(adoc)
            if len(doc) == 1:
                # we'll just return the single doc
                # otherwise it's up to the caller to do something with that
                doc = doc.pop()
        assert doc, "No document could be found"
        return doc

    @property
    def collection(self):
        """
        Return collection object used to fetch doc in which we store status
        """
        raise NotImplementedError("implement me in sub-class")

    def register_status(self, doc, stage, status, transient=False, init=False, **extra):
        assert self.collection, "No collection set"
        # stage: "snapshot", "publish", etc... depending on the what's being done
        job_info = {
            'status': status,
            'step_started_at': datetime.datetime.now(),
            'logfile': self.logfile
        }
        stage_info = {}
        stage_key = None
        # register status can be about different stages:
        stage_info.setdefault(stage, {}).update(extra[stage])
        stage_key = list(extra[stage].keys())
        assert len(stage_key) == 1, stage_key
        stage_key = stage_key.pop()
        if transient:
            # record some "in-progress" information
            job_info['pid'] = os.getpid()
        else:
            # only register time when it's a final state
            job_info["time"] = timesofar(self.ti)
            t1 = round(time.time() - self.ti, 0)
            job_info["time_in_s"] = t1
            stage_info.setdefault(stage, {}).setdefault(stage_key, {}).update({"created_at": datetime.datetime.now()})
        if "job" in extra:
            job_info.update(extra["job"])
        # since the base is the merged collection, we register info there
        if init:
            # init timer for this step
            self.ti = time.time()
            self.collection.update({'_id': doc["_id"]}, {"$push": {'jobs': job_info}})
            # now refresh/sync
            doc = self.collection.find_one({'_id': doc["_id"]})
        else:
            # merge extra at root level
            doc["jobs"] and doc["jobs"].append(job_info)

            def merge_index_info(target, d):
                if not isinstance(target, dict):
                    # previous value wasn't a dict, just replace
                    target = d
                elif "__REPLACE__" in d.keys():
                    d.pop("__REPLACE__")
                    target = d
                else:
                    if status == "success":
                        # remove 'err' key to avoid merging success results with errors
                        target.pop("err", None)
                    for k, v in d.items():
                        if isinstance(v, dict):
                            if k in target:
                                target[k] = merge_index_info(target[k], v)
                            else:
                                v.pop("__REPLACE__", None)
                                # merge v with "nothing" just to make sure to remove any "__REPLACE__"
                                v = merge_index_info({}, v)
                                target[k] = v
                        else:
                            target[k] = v
                return target

            doc = merge_index_info(doc, stage_info)
            self.collection.replace_one({"_id": doc["_id"]}, doc)


class BaseSourceManager(BaseManager):
    """
    Base class to provide source management: discovery, registration
    Actual launch of tasks must be defined in subclasses
    """

    # define the class manager will look for. Set in a subclass
    SOURCE_CLASS = None

    def __init__(self, job_manager, datasource_path="dataload.sources", *args, **kwargs):
        super(BaseSourceManager, self).__init__(job_manager, *args, **kwargs)
        self.conn = get_src_conn()
        self.default_src_path = datasource_path
        self.poll_schedule = None

    def filter_class(self, klass):
        """
        Gives opportunity for subclass to check given class and decide to
        keep it or not in the discovery process. Returning None means "skip it".
        """
        # keep it by default
        return klass

    def register_classes(self, klasses):
        """
        Register each class in self.register dict. Key will be used
        to retrieve the source class, create an instance and run method from it.
        It must be implemented in subclass as each manager may need to access
        its sources differently,based on different keys.
        """
        raise NotImplementedError("implement me in sub-class")

    def find_classes(self, src_module, fail_on_notfound=True):
        """
        Given a python module, return a list of classes in this module, matching
        SOURCE_CLASS (must inherit from)
        """
        # try to find a uploader class in the module
        klasses = []
        for attr in dir(src_module):
            something = getattr(src_module, attr)
            # not interested in classes coming from biothings.hub.*, these would typically come
            # from "from biothings.hub.... import aclass" statements and would be incorrectly registered
            # we only look for classes defined straight from the actual module
            if type(something) == type and issubclass(something, self.__class__.SOURCE_CLASS) and \
                    not something.__module__.startswith("biothings.hub"):
                klass = something
                if not self.filter_class(klass):
                    continue
                logger.debug("Found a class based on %s: '%s'", self.__class__.SOURCE_CLASS.__name__, klass)
                klasses.append(klass)
        if not klasses:
            if fail_on_notfound:
                raise UnknownResource("Can't find a class based on %s in module '%s'" % (self.__class__.SOURCE_CLASS.__name__, src_module))
        return klasses

    def register_source(self, src, fail_on_notfound=True):
        """Register a new data source. src can be a module where some classes
        are defined. It can also be a module path as a string, or just a source name
        in which case it will try to find information from default path.
        """
        if isinstance(src, str):
            try:
                src_m = importlib.import_module(src)
                src_m = importlib.reload(src_m)
            except ImportError:
                try:
                    src_m = importlib.import_module("%s.%s" % (self.default_src_path, src))
                except ImportError:
                    msg = "Can't find module '%s', even in '%s'" % (src, self.default_src_path)
                    logger.error(msg)
                    raise UnknownResource(msg)

        elif isinstance(src, dict):
            # source is comprised of several other sub sources
            assert len(src) == 1, "Should have only one element in source dict '%s'" % src
            _, sub_srcs = list(src.items())[0]
            for src in sub_srcs:
                self.register_source(src, fail_on_notfound)
            return
        else:
            src_m = src
        # first try to find classes defined in __init__.py (in package) explicitely
        klasses = self.find_classes(src_m, fail_on_notfound)
        # then if none found, try to search within the package's modules
        if not klasses:
            try:
                src_m_path = src_m.__path__[0]
                for d in os.listdir(src_m_path):
                    if d.endswith("__pycache__"):
                        continue
                    modpath = os.path.join(src_m.__name__, d).replace(".py", "").replace(os.path.sep, ".")
                    try:
                        m = importlib.import_module(modpath)
                        klasses.extend(self.find_classes(m, fail_on_notfound))
                    except (SyntaxError, ImportError) as e:
                        logger.debug("Couldn't import %s: %s", modpath, e)
                        continue
            except TypeError as e:
                logger.warning("Can't register source '%s', something's wrong with path: %s", src_m, e)
        logger.debug("Found classes to register: %s", repr(klasses))

        self.register_classes(klasses)

    def register_sources(self, sources):
        assert not isinstance(sources, str), "sources argument is a string, should pass a list"
        self.register.clear()
        for src in sources:
            try:
                # batch registration, we'll silently ignore not-found sources
                self.register_source(src, fail_on_notfound=False)
            except (UnknownResource, ResourceError) as e:
                logger.info("Can't register source '%s', skip it; %s", src, e)
                import traceback
                logger.error(traceback.format_exc())


class JobManager(object):
    # TODO: Add class docstring
    COLUMNS = ["pid", "source", "category", "step", "description", "mem", "cpu", "started_at", "duration"]
    HEADER = dict(zip(COLUMNS, [c.upper() for c in COLUMNS]))  # upper() for column titles
    HEADERLINE = "{pid:^10}|{source:^35}|{category:^10}|{step:^20}|{description:^30}|{mem:^10}|{cpu:^6}|{started_at:^20}|{duration:^10}"
    DATALINE = HEADERLINE.replace("^", "<")

    def __init__(self, loop, process_queue=None, thread_queue=None, max_memory_usage=None,
                 num_workers=None, num_threads=None, default_executor="thread", auto_recycle=True):
        if not os.path.exists(config.RUN_DIR):
            logger.info("Creating RUN_DIR directory '%s'", config.RUN_DIR)
            os.makedirs(config.RUN_DIR)
        self.loop = loop
        self.num_workers = num_workers
        if self.num_workers == 0:
            logger.debug("Adjusting number of worker to 1")
            self.num_workers = 1
        self.num_threads = num_threads or self.num_workers
        self.process_queue = process_queue or concurrent.futures.ProcessPoolExecutor(max_workers=self.num_workers)
        # TODO: limit the number of threads (as argument) ?
        self.thread_queue = thread_queue or concurrent.futures.ThreadPoolExecutor(max_workers=self.num_threads)
        if default_executor == "thread":
            self.loop.set_default_executor(self.thread_queue)
        else:
            self.loop.set_default_executor(self.process_queue)
        self.ok_to_run = asyncio.Semaphore()
        # auto-creata RUN_DIR
        if not os.path.exists(config.RUN_DIR):
            os.makedirs(config.RUN_DIR)

        if max_memory_usage == "auto":
            # try to find a nice limit...
            limited = int(psutil.virtual_memory().available * .6)
            logger.info("Auto-setting memory usage limit to %s", sizeof_fmt(limited))
            max_memory_usage = limited
        elif max_memory_usage:
            logger.info("Setting memory usage to %s", sizeof_fmt(max_memory_usage))
        else:
            logger.info("No memory limit set")
        self.max_memory_usage = max_memory_usage
        self.avail_memory = int(psutil.virtual_memory().available)
        self._phub = None
        self.auto_recycle = auto_recycle  # active
        self.auto_recycle_setting = auto_recycle  # keep setting if we need to restore it its orig value
        self.jobs = {}  # all active jobs (thread/process)
        self._pchildren = []
        self.clean_staled()

    def stop(self, force=False, recycling=False, wait=1):
        @asyncio.coroutine
        def do():
            try:
                # shutting down the process queue can take a while
                # if some processes are still running (it'll wait until they're done)
                # we'll wait in a thread to prevent the hub from being blocked
                logger.info("Shutting down current process queue...")
                pinfo = {"__skip_check__": True,  # skip sanity check, mem check to make sure
                                                  # this worker will be run
                         "category": "admin",
                         "source": "maintenance",
                         "step": "",
                         "description": "Stopping process queue"}
                j = yield from self.defer_to_thread(pinfo, self.process_queue.shutdown)
                yield from j
                if recycling:
                    # now replace
                    logger.info("Replacing process queue with new one")
                    self.process_queue = concurrent.futures.ProcessPoolExecutor(max_workers=self.num_workers)
                else:
                    self.process_queue = None
            except Exception as e:
                logger.error("Error while recycling the process queue: %s", e)
                raise

        @asyncio.coroutine
        def kill():
            nonlocal wait
            if wait < 1:
                wait = 1  # wait a little bit so job manager has time to stop if nothing is running
            logger.warning("Wait %s seconds before killing queue processes", wait)
            yield from asyncio.sleep(wait)
            logger.warning("Can't wait anymore, killing running processed in the queue !")
            for proc in self.pchildren:
                logger.warning("Killing %s", proc)
                proc.kill()

        def done(f):
            f.result()  # consume future's result to potentially raise exception
        fut = asyncio.ensure_future(do())
        fut.add_done_callback(done)
        if force:
            futkill = asyncio.ensure_future(kill())
        return fut

    def clean_staled(self):
        # clean old/staled files
        children_pids = [p.pid for p in self.pchildren]
        active_tids = [t.getName() for t in self.thread_queue._threads]
        pid_pat = re.compile(".*/(\d+)_.*\.pickle")  # see track() for filename format
        for fn in glob.glob(os.path.join(config.RUN_DIR, "*.pickle")):
            pid = pid_pat.findall(fn)
            if not pid:
                continue
            try:
                pid = int(pid[0].split("_")[0])
            except IndexError:
                logger.warning("Invalid PID file '%s', skip it", fn)
                raise
            if pid not in children_pids:
                logger.info("Removing staled pid file '%s'", fn)
                os.unlink(fn)
        tid_pat = re.compile(".*/(Thread\w*-\d+)_.*\.pickle")
        for fn in glob.glob(os.path.join(config.RUN_DIR, "*.pickle")):
            try:
                tid = tid_pat.findall(fn)[0].split("_")[0]
            except IndexError:
                logger.warning("Invalid TID file '%s', skip it", fn)
                raise
            if not tid:
                continue
            if tid not in active_tids:
                logger.info("Removing staled thread file '%s'", fn)
                os.unlink(fn)

    def recycle_process_queue(self):
        """
        Replace current process queue with a new one. When processes
        are used over and over again, memory tends to grow as python
        interpreter keeps some data (...). Calling this method will
        perform a clean shutdown on current queue, waiting for running
        processes to terminate, then discard current queue and replace
        it a new one.
        """
        return self.stop(recycling=True)

    @asyncio.coroutine
    def check_constraints(self, pinfo=None):
        mem_req = pinfo and pinfo.get("__reqs__", {}).get("mem") or 0
        t0 = time.time()
        waited = False
        sleep_time = 5
        if mem_req:
            logger.info(
                "Job {cat:%s,source:%s,step:%s} requires %s memory, checking if available",
                pinfo.get("category"), pinfo.get("source"), pinfo.get("step"), sizeof_fmt(mem_req))
        if self.max_memory_usage:
            hub_mem = self.hub_memory
            while hub_mem >= self.max_memory_usage:
                if self.auto_recycle:
                    pworkers = self.get_pid_files()
                    tworkers = self.get_thread_files()
                    if len(pworkers) == 0 and len(tworkers) == 0:
                        logger.info("No worker running, recycling the process queue...")
                        fut = self.recycle_process_queue()

                        def recycled(f):
                            res = f.result()
                            # still out of memory ?
                            avail_mem = self.max_memory_usage - self.hub_memory
                            if avail_mem <= 0:
                                logger.error(
                                    "After recycling process queue, "
                                    "memory usage is still too high (needs at least %s more)"
                                    "now turn auto-recycling off to prevent infinite recycling...",
                                    sizeof_fmt(abs(avail_mem))
                                )
                                self.auto_recycle = False

                        fut.add_done_callback(recycled)
                logger.info(
                    "Hub is using too much memory to launch job {cat:%s,source:%s,step:%s}"\
                    " (%s used, more than max allowed %s), wait a little (job's already been postponed for %s)",
                    pinfo.get("category"), pinfo.get("source"), pinfo.get("step"),
                    sizeof_fmt(hub_mem), sizeof_fmt(self.max_memory_usage), timesofar(t0)
                )
                yield from asyncio.sleep(sleep_time)
                waited = True
                hub_mem = self.hub_memory
        if mem_req:
            # max allowed mem is either the limit we gave and the os limit
            max_mem = self.max_memory_usage and self.max_memory_usage or self.avail_memory
            # TODO: check projected memory (jobs with mem requirements currently running
            # as those jobs may not have reached their max mem usage yet)
            hub_mem = self.hub_memory
            while mem_req >= (max_mem - hub_mem):
                logger.info(
                    "Job {cat:%s,source:%s,step:%s} needs %s to run, not enough to launch it "
                    "(hub consumes %s while max allowed is %s), wait a little  (job's already been postponed for %s)",
                    pinfo.get("category"), pinfo.get("source"), pinfo.get("step"), sizeof_fmt(mem_req),
                    sizeof_fmt(hub_mem), sizeof_fmt(max_mem), timesofar(t0)
                )
                yield from asyncio.sleep(sleep_time)
                waited = True
                # refresh limites and usage (manager can be modified from hub
                # thus memory usage can be modified on-the-fly
                hub_mem = self.hub_memory
                max_mem = self.max_memory_usage and self.max_memory_usage or self.avail_memory
        pendings = len(self.process_queue._pending_work_items.keys()) - config.HUB_MAX_WORKERS
        while pendings >= config.MAX_QUEUED_JOBS:
            if not waited:
                logger.info(
                    "Can't run job {cat:%s,source:%s,step:%s} right now, too much pending jobs in the queue (max: %s), will retry until possible",
                    pinfo.get("category"), pinfo.get("source"), pinfo.get("step"), config.MAX_QUEUED_JOBS
                )
            yield from asyncio.sleep(sleep_time)
            pendings = len(self.process_queue._pending_work_items.keys()) - config.HUB_MAX_WORKERS
            waited = True
        # finally check custom predicates
        predicates = pinfo and pinfo.get("__predicates__", [])
        failed_predicate = None
        while True:
            for predicate in predicates:
                if not predicate(self):
                    failed_predicate = predicate
                    break  # for loop (most inner one)
                else:
                    # reset flag
                    failed_predicate = None
            if failed_predicate:
                logger.info("Can't run job {cat:%s,source:%s,step:%s} right now, predicate %s failed, will retry until possible",
                            pinfo.get("category"), pinfo.get("source"), pinfo.get("step"), failed_predicate)
                yield from asyncio.sleep(sleep_time)
                waited = True
            else:
                break  # while loop
        if waited:
            logger.info("Job {cat:%s,source:%s,step:%s} now can be launched (total waiting time: %s)",
                        pinfo.get("category"), pinfo.get("source"), pinfo.get("step"), timesofar(t0))
            # auto-recycle could have been temporarily disabled until more mem is assigned.
            # if we've been able to run the job, it means we had enough mem so restore
            # recycling setting (if auto_recycle was False, it's ignored
            if self.auto_recycle_setting:
                self.auto_recycle = self.auto_recycle_setting

    @asyncio.coroutine
    def defer_to_process(self, pinfo=None, func=None, *args):

        @asyncio.coroutine
        def run(future, job_id):
            nonlocal pinfo
            yield from self.check_constraints(pinfo)
            self.ok_to_run.release()
            # pinfo can contain predicates hardly pickleable during run_in_executor
            # but we also need not to touch the original one
            copy_pinfo = copy.deepcopy(pinfo)
            copy_pinfo.pop("__predicates__", None)
            self.jobs[job_id] = copy_pinfo
            res = self.loop.run_in_executor(
                self.process_queue,
                partial(do_work, job_id, "process", copy_pinfo, func, *args)
            )

            def ran(f):
                try:
                    # consume future, just to trigger potential exceptions
                    r = f.result()
                finally:
                    # whatever the result we want to make sure to clean the job registry
                    # to keep it sync with actual running jobs
                    self.jobs.pop(job_id)
            res.add_done_callback(ran)
            res = yield from res
            # process could generate other parallelized jobs and return a Future/Task
            # If so, we want to make sure we get the results from that task
            if type(res) == asyncio.Task:
                res = yield from res
            future.set_result(res)
        yield from self.ok_to_run.acquire()
        f = asyncio.Future()

        def runned(innerf, job_id):
            if innerf.exception():
                f.set_exception(innerf.exception())
        job_id = get_random_string()
        fut = asyncio.ensure_future(run(f, job_id))
        fut.add_done_callback(partial(runned, job_id=job_id))
        return f

    @asyncio.coroutine
    def defer_to_thread(self, pinfo=None, func=None, *args):

        skip_check = pinfo.get("__skip_check__", False)

        @asyncio.coroutine
        def run(future, job_id):
            if not skip_check:
                yield from self.check_constraints(pinfo)
                self.ok_to_run.release()
            self.jobs[job_id] = pinfo
            res = self.loop.run_in_executor(
                self.thread_queue,
                partial(do_work, job_id, "thread", pinfo, func, *args)
            )

            def ran(f):
                try:
                    r = f.result()
                finally:
                    # whatever the result we want to make sure to clean the job registry
                    # to keep it sync with actual running jobs
                    self.jobs.pop(job_id)
            res.add_done_callback(ran)
            res = yield from res
            # thread could generate other parallelized jobs and return a Future/Task
            # If so, we want to make sure we get the results from that task
            if type(res) == asyncio.Task:
                res = yield from res
            future.set_result(res)
        if not skip_check:
            yield from self.ok_to_run.acquire()
        f = asyncio.Future()

        def runned(innerf, job_id):
            if innerf.exception():
                f.set_exception(innerf.exception())
        job_id = get_random_string()
        fut = asyncio.ensure_future(run(f, job_id))
        fut.add_done_callback(partial(runned, job_id=job_id))
        return f

    def submit(self, pfunc, schedule=None):
        """
        Helper to submit and run tasks. Tasks will run async'ly.
        pfunc is a functools.partial
        schedule is a string representing a cron schedule, task will then be scheduled
        accordingly.
        """
        logger.info("Building task: %s", pfunc)
        if schedule:
            logger.info("Scheduling task %s: %s", pfunc, schedule)
            cron = aiocron.crontab(schedule, func=pfunc, start=True, loop=self.loop)
            return cron
        else:
            ff = asyncio.ensure_future(pfunc())
            return ff

    def schedule(self, crontab, func, *args, **kwargs):
        """
        Helper to create a cron job from a callable "func". *argd, and **kwargs
        are passed to func. "crontab" follows aicron notation.
        """
        # we need to dynamically create a wrapper coroutine with a name
        # that makes sense, taken from func, otherwise all scheduled jobs would
        # have the same wrapping coroutine name
        if isinstance(func,partial):
            func_name = func.func.__name__
        else:
            func_name = func.__name__
        strcode = """
@asyncio.coroutine
def %s():
    func(*args, **kwargs)
""" % func_name
        code = compile(strcode, "<string>", "exec")
        command_globals = {}
        command_locals = {
            "asyncio": asyncio,
            "func": func,
            "args": args,
            "kwargs": kwargs
        }
        eval(code, command_locals, command_globals)
        run_func = command_globals[func_name]
        job = self.submit(run_func,schedule=crontab)

        return job

    @property
    def hub_process(self):
        if not self._phub:
            self._phub = find_process(os.getpid())
        return self._phub

    @property
    def pchildren(self):
        if not self._pchildren:
            self._pchildren = self.hub_process.children()
        return self._pchildren

    @property
    def hub_memory(self):
        total_mem = 0
        try:
            procs = [self.hub_process] + self.pchildren
            for proc in procs:
                total_mem += proc.memory_info().rss
        except psutil.NoSuchProcess:
            # observed multiple time: hub main pid doesn't exist, like it was replace, not sure why,... OS ?
            self._phub = None
            self._pchildren = None

        return total_mem

    def get_pid_files(self, child=None):
        pids = {}
        try:
            pat = re.compile(".*/(\d+)_.*\.pickle")  # see track() for filename format
            children_pids = [p.pid for p in self.pchildren]
            for fn in glob.glob(os.path.join(config.RUN_DIR, "*.pickle")):
                try:
                    pid = int(pat.findall(fn)[0].split("_")[0])
                    if not child or child.pid == pid:
                        try:
                            worker = pickle.load(open(fn, "rb"))
                        except FileNotFoundError:
                            # it's possible that, as this point, the pickle file
                            # doesn't exist anymore (process is done and file was unlinked)
                            # just ignore go to next one
                            continue
                        proc = self.pchildren[children_pids.index(pid)]

                        worker["process"] = {
                            "mem": proc.memory_info().rss,
                            "cpu": proc.cpu_percent()
                        }
                        pids[pid] = worker
                except IndexError:
                    # weird though... should have only pid files there...
                    pass
        finally:
            return pids

    def get_thread_files(self):
        tids = {}
        try:
            # see track() for filename format
            pat = re.compile(".*/(Thread\w*-\d+)_.*\.pickle")
            threads = self.thread_queue._threads
            active_tids = [t.getName() for t in threads]
            for fn in glob.glob(os.path.join(config.RUN_DIR, "*.pickle")):
                try:
                    tid = pat.findall(fn)[0].split("_")[0]
                    worker = pickle.load(open(fn, "rb"))
                    worker["process"] = self.hub_process  # misleading... it's the hub process
                    tids[tid] = worker
                except IndexError:
                    # weird though... should have only pid files there...
                    pass
        finally:
            return tids

    def extract_pending_info(self, pending):
        info = pending.fn.args[2]
        assert isinstance(info, dict)
        return info

    def extract_worker_info(self, worker):
        info = OrderedDict()
        proc = worker.get("process", worker)
        err = worker.get("err") and " !" or ""
        info["pid"] = str(worker["job"]["id"]) + err
        info["source"] = norm(worker["job"].get("source") or "", 25)
        info["category"] = norm(worker["job"].get("category") or "", 10)
        info["step"] = norm(worker["job"].get("step") or "", 20)
        info["description"] = norm(worker["job"].get("description") or "", 30)
        info["mem"] = sizeof_fmt(proc.get("memory", {}).get("size", 0.0))
        info["cpu"] = "%.1f%%" % proc.get("cpu", {}).get("percent", 0.0)
        info["started_at"] = worker["job"]["started_at"]
        if worker.get("duration"):
            info["duration"] = worker["duration"]
        else:
            info["duration"] = timesofar(worker["job"]["started_at"])
        # for now, don't display files used by the process
        info["files"] = []
        #if proc:
        #    for pfile in proc.open_files():
        #        # skip 'a' (logger)
        #        if pfile.mode == 'r':
        #            finfo = OrderedDict()
        #            finfo["path"] = pfile.path
        #            finfo["read"] = sizeof_fmt(pfile.position)
        #            size = os.path.getsize(pfile.path)
        #            finfo["size"] = sizeof_fmt(size)
        #            #info["files"].append(finfo)
        return info

    def print_workers(self, workers):
        if workers:
            out = []
            out.append(self.__class__.HEADERLINE.format(**self.__class__.HEADER))
            for pid in workers:
                worker = workers[pid]
                info = self.extract_worker_info(worker)
                tt = datetime.datetime.fromtimestamp(info["started_at"]).timetuple()
                info["started_at"] = time.strftime("%Y/%m/%d %H:%M:%S", tt)
                try:
                    out.append(self.__class__.DATALINE.format(**info))
                except (TypeError, KeyError) as e:
                    out.append(e)
                    out.append(pformat(info))

            return "\n".join(out)
        else:
            return ""

    def print_pending_info(self, num, info):
        assert isinstance(info, dict)
        info["cpu"] = ""
        info["mem"] = ""
        info["pid"] = ""
        info["duration"] = ""
        info["source"] = norm(info["source"], 35)
        info["category"] = norm(info["category"], 10)
        info["step"] = norm(info["step"], 20)
        info["description"] = norm(info["description"], 30)
        info["started_at"] = ""
        out = []
        try:
            out.append(self.__class__.DATALINE.format(**info))
        except (TypeError, KeyError) as e:
            out.append(e)
            out.append(pformat(info))

        return out

    def get_process_summary(self):
        running_pids = self.get_pid_files()
        res = {}
        for child in self.pchildren:
            try:
                mem = child.memory_info().rss
                pio = child.io_counters()
                # TODO: cpu as reported here isn't reliable, the only to get something
                # consistent to call cpu_percent() with a waiting time argument to integrate
                # CPU activity over this time, but this is a blocking call and freeze the hub
                # (an async implementation might possible though). Currently, pchildren is list
                # set at init time where process object are stored, so subsequent cpu_percent()
                # calls should report CPU activity since last call (between /job_manager & top()
                # calls), but it constently return CPU > 100% even when no thread running (that
                # could have been the explination but it's not).
                cpu = child.cpu_percent()
                res[child.pid] = {
                    "memory": {
                        "size": child.memory_info().rss,
                        "percent": child.memory_percent(),
                    },
                    "cpu": {
                        # override status() when we have cpu activity to avoid
                        # having a "sleeping" process that's actually running something
                        # (prob happening because delay between status and cpu_percent(), like a race condition)
                        "status": cpu > 0.0 and "running" or child.status(),
                        "percent": cpu
                    },
                    "io": {
                        "read_count": pio.read_count,
                        "write_count": pio.write_count,
                        "read_bytes": pio.read_bytes,
                        "write_bytes": pio.write_bytes
                    }
                }

                if child.pid in running_pids:
                    # something is running on that child process
                    worker = running_pids[child.pid]
                    res[child.pid]["job"] = {
                        "started_at": worker["job"]["started_at"],
                        "duration": timesofar(worker["job"]["started_at"], 0),
                        "func_name": worker["func_name"],
                        "category": worker["job"]["category"],
                        "description": worker["job"]["description"],
                        "source": worker["job"]["source"],
                        "step": worker["job"]["step"],
                        "id": worker["job"]["id"],
                    }
            except psutil.NoSuchProcess as e:
                print("child not found %s %s" % (child, e))
                continue

        return res

    def get_thread_summary(self):
        running_tids = self.get_thread_files()
        tchildren = self.thread_queue._threads
        res = {}
        for child in tchildren:
            res[child.name] = {
                "is_alive": child.isAlive(),
                "is_daemon": child.isDaemon(),
            }

            if child.name in running_tids:
                # something is running on that child process
                worker = running_tids[child.name]
                res[child.name]["job"] = {
                    "started_at": worker["job"]["started_at"],
                    "duration": timesofar(worker["job"]["started_at"], 0),
                    "func_name": worker["func_name"],
                    "category": worker["job"]["category"],
                    "description": worker["job"]["description"],
                    "source": worker["job"]["source"],
                    "step": worker["job"]["step"],
                    "id": worker["job"]["id"],
                }

        return res

    def get_summary(self, child=None):
        pworkers = self.get_pid_files(child)
        tworkers = self.get_thread_files()
        ppendings = self.get_pending_processes()
        tpendings = {}  # TODO:
        return {
            "process": {
                "running": list(pworkers.keys()),
                "pending": list(ppendings.keys()),
                "all": self.get_process_summary(),
                "max": self.process_queue._max_workers
            },
            "thread": {
                "running": list(tworkers.keys()),
                "pending": list(tpendings.keys()),
                "all": self.get_thread_summary(),
                "max": self.thread_queue._max_workers,
            },
            "memory": self.hub_memory,
            "available_system_memory": self.avail_memory,
            "max_memory_usage": self.max_memory_usage,
            "hub_pid": self.hub_process.pid
        }

    def get_pending_summary(self, getstr=False):
        running = len(self.get_pid_files())
        return "%d pending job(s)" % (len(self.process_queue._pending_work_items) - running)

    def get_pending_processes(self):
        # pendings are kept in queue while running, until result is there so we need
        # to adjust the actual real pending jobs. also, pending job are get() from the
        # queue following FIFO order. finally, worker ID is incremental. So...
        pendings = sorted(self.process_queue._pending_work_items.items())
        running = len(self.get_pid_files())
        actual_pendings = dict(pendings[running:])
        return actual_pendings

    def show_pendings(self, running=None):
        out = []
        out.append(self.get_pending_summary())
        actual_pendings = self.get_pending_processes()
        if actual_pendings:
            out.append(self.__class__.HEADERLINE.format(**self.__class__.HEADER))
            for num, pending in actual_pendings.items():
                info = self.extract_pending_info(pending)
                try:
                    self.print_pending_info(num, info)
                except Exception as e:
                    out.append(e)
                    out.append(pformat(pending))

        return "\n".join(out)

    def top(self, action="summary"):
        pending = False
        done = False
        run = False
        pid = None
        child = None
        #if action:
        #    try:
        #        # want to see details for a specific process ?
        #        pid = int(action)
        #        child = [p for p in pchildren if p.pid == pid][0]
        #    except ValueError:
        #        pass
        pworkers = self.get_pid_files(child)
        tworkers = self.get_thread_files()
        done_jobs = glob.glob(os.path.join(config.RUN_DIR, "done", "*.pickle"))
        out = []
        if child:
            return pworkers[child.pid]
        elif action == "pending":
            return self.show_pendings(running=len(pworkers))
        elif action == "summary":
            res = self.get_summary()
            # TODO: delete the follow two lines
            # pworkers = dict([(pid, proc) for pid, proc in res["process"]["all"].items() if pid in res["process"]["running"]])
            # tworkers = dict([(tid, thread) for tid, thread in res["thread"]["all"].items() if tid in res["thread"]["running"]])
            pworkers = {pid: proc for pid, proc in res["process"]["all"].items() if pid in res["process"]["running"]}
            tworkers = {tid: thread for tid, thread in res["thread"]["all"].items() if tid in res["thread"]["running"]}

            out.append(self.print_workers(pworkers))
            out.append(self.print_workers(tworkers))
            out.append("%d running job(s)" % (len(pworkers) + len(tworkers)))
            out.append("%s, type 'top(pending)' for more" % self.get_pending_summary())
            if done_jobs:
                out.append("%s finished job(s), type 'top(done)' for more" % len(done_jobs))
        else:
            raise ValueError("Unknown action '%s'" % action)

        return "\n".join(out)

    def job_info(self):
        summary = self.get_summary()
        prunning = summary["process"]["running"]
        trunning = summary["thread"]["running"]
        ppending = summary["process"]["pending"]
        tpending = summary["thread"]["pending"]
        return {
            "queue": {
                "process": summary["process"],
                "thread": summary["thread"],
            },
            "memory": summary["memory"],
            "available_system_memory": summary["available_system_memory"],
            "max_memory_usage": summary["max_memory_usage"],
            "hub_pid": summary["hub_pid"]
        }

# just a helper to clean/prepare job's values printing
def norm(value, maxlen):
    if len(value) > maxlen:
        value = "...%s" % value[-maxlen+3:]
    return value



from biothings.utils.dataload import alwayslist


class BaseMapper(object):
    """
    Basic mapper used to convert documents.
    if mapper's name matches source's metadata's mapper,
    mapper.convert(docs) call will be used to
    process/convert/whatever passed documents
    """

    def __init__(self, name=None, *args, **kwargs):
        self.name = name

    def load(self):
        """
        Do whatever is required to fill mapper with mapping data
        Can be called multiple time, the first time only will load data
        """
        raise NotImplementedError("sub-class and implement me")

    def process(self, docs):
        """
        Convert given docs into other docs.
        """
        raise NotImplementedError("sub-class and implement me")


class IDBaseMapper(BaseMapper):
    """
    Provide mapping between different sources
    """

    def __init__(self, name=None, convert_func=None, *args, **kwargs):
        """
        'name' may match a "mapper" metatdata field (see uploaders). If None, mapper
        will be applied to any document from a resource without "mapper" argument
        """
        super(IDBaseMapper, self).__init__(name=name)
        self.map = None
        self.convert_func = convert_func

    def translate(self, _id, transparent=False):
        """
        Return _id translated through mapper, or _id itself if not part of mapper
        If 'transparent' and no match, original _id will be returned
        """
        if self.need_load():
            self.load()
        default = transparent and _id or None
        conv = self.convert_func or (lambda x: x)
        return self.map.get(conv(_id), default)

    def __contains__(self, _id):
        if self.need_load():
            self.load()
        return _id in self.map

    def __len__(self):
        if self.need_load():
            self.load()
        return len(self.map)

    def process(self, docs, key_to_convert="_id", transparent=True):
        """
        Process 'key_to_convert' document key using mapping.
        If transparent and no match, original key will be used
        (so there's no change). Else, if no match, document will
        be discarded (default).
        Warning: key to be translated must not be None (it's considered
        a non-match)
        """
        for doc in docs:
            _id = doc.get(key_to_convert)
            _newid = self.translate(_id, transparent)
            if _newid is None and not transparent:
                continue
            for _oneid in alwayslist(_newid):
                _oneid = str(_oneid)
                doc[key_to_convert] = _oneid
                yield doc

    def need_load(self):
        return self.map is None


class TransparentMapper(BaseMapper):

    def load(self, *args, **kwargs):
        pass

    def process(self, docs, *args, **kwargs):
        return docs



import biothings.utils.redis as redis
from biothings.utils.mongo import doc_feeder
import pymongo


class BasePreCompiledDataProvider(object):
    def __init__(self, name):
        """
        'name' is a way to identify this provider
        (usually linked to a database name behind the scene)
        """
        self.name = name

    def register(self, _id, col_name):
        """
        Tell provider that _id can be found in collection named 'col_name'
        """
        raise NotImplementedError("implement in sub-class")

    def get_all(self):
        """
        Iterate over all register _ids, return a list of collection names
        where they can be found
        """
        raise NotImplementedError("implement in sub-class")


class RedisPreCompiledDataProvider(BasePreCompiledDataProvider):
    def __init__(self, name, connection_params):
        super(RedisPreCompiledDataProvider, self).__init__(name)
        self.connection_params = connection_params
        self.client = redis.RedisClient(connection_params)
        try:
            self.client.check()
        except AssertionError:
            self.client.initialize()
        self.db = self.client.get_db(self.name)

    def register(self, _id, col_name):
        self.db.hset(_id, col_name, 1)

    def get_all(self):
        for _id in self.db.scan_iter():
            #cols = list(self.db.hgetall(_id).keys())
            cols = []
            yield (_id, cols)


class MongoDBPreCompiledDataProvider(BasePreCompiledDataProvider):
    def __init__(self, db_name, name, connection_params):
        self.db_name = db_name
        self.col_name = name
        self.connection_params = connection_params
        self.client = pymongo.MongoClient(connection_params)
        self.col = self.client[self.db_name][self.col_name]

    def register(self, _id, col_name):
        updt = {"$set": {"srcs.%s" % col_name: 1}}
        if type(_id) == list:
            bob = self.col.initialize_unordered_bulk_op()
            for oneid in _id:
                bob.find({"_id": oneid}).upsert().update(updt)
            bob.execute()
        else:
            self.col.update_one({"_id": _id}, updt, upsert=True)

    def get_all(self, batch_size=100000):
        for doc_ids in doc_feeder(self.col, step=batch_size, inbatch=True):
            for d in doc_ids:
                yield d



''' For Google Analytics tracking in web '''
from tornado.httpclient import HTTPRequest, AsyncHTTPClient
import re
from random import randint
from operator import itemgetter
from urllib.parse import quote_plus as _q

RE_LOCALE = re.compile(r'(^|\s*,\s*)([a-zA-Z]{1,8}(-[a-zA-Z]{1,8})*)\s*(;\s*q\s*=\s*(1(\.0{0,3})?|0(\.[0-9]{0,3})))?', re.I)

def get_user_language(lang):
    user_locals = []
    matched_locales = RE_LOCALE.findall(str(lang))
    if matched_locales:
        lang_lst = map((lambda x: x.replace('-', '_')), (i[1] for i in matched_locales))
        quality_lst = map((lambda x: x and x or 1), (float(i[4] and i[4] or '0') for i in matched_locales))
        lang_quality_map = map((lambda x, y: (x, y)), lang_lst, quality_lst)
        user_locals = [x[0] for x in sorted(lang_quality_map, key=itemgetter(1), reverse=True)]

    if user_locals:
        return user_locals[0]
    else:
        return ''

def generate_hash(user_agent, screen_resolution, screen_color_depth):
    tmpstr = "%s%s%s" % (user_agent, screen_resolution, screen_color_depth)
    hash_val = 1

    if tmpstr:
        hash_val = 0
        for ordinal in map(ord, tmpstr[::-1]):
            hash_val = ((hash_val << 6) & 0xfffffff) + ordinal + (ordinal << 14)
            left_most_7 = hash_val & 0xfe00000
            if left_most_7 != 0:
                hash_val ^= left_most_7 >> 21

    return hash_val

def generate_unique_id(user_agent='', screen_resolution='', screen_color_depth=''):
    '''Generates a unique user ID from the current user-specific properties.'''
    return ((randint(0, 0x7fffffff) ^ generate_hash(user_agent, screen_resolution, screen_color_depth))
            & 0x7fffffff)

# This is a mixin for biothing handlers, and references class variables from that class, cannot be used
# without mixing in
class GAMixIn:
    def ga_track(self, event={}):
        no_tracking = self.get_argument('no_tracking', None)
        is_prod = not self.settings.get('debug', False)
        if not no_tracking and is_prod and self.web_settings.GA_ACCOUNT:
            _req = self.request
            path = _req.path
            ln = _req.headers.get('Accept-Language', '')
            remote_ip = _req.headers.get("X-Real-Ip", _req.headers.get("X-Forwarded-For", _req.remote_ip))
            user_agent = _req.headers.get("User-Agent", "")
            host = _req.headers.get("Host", "N/A")
            this_user = generate_unique_id(user_agent=user_agent)
            user_agent = _q(user_agent)
            langua = get_user_language(ln)
            # compile measurement protocol string for google
            # first do the pageview hit type
            request_body = 'v=1&t=pageview&tid={}&ds=web&cid={}&uip={}&ua={}&an={}&av={}&dh={}&dp={}'.format(
                self.web_settings.GA_ACCOUNT, this_user, remote_ip, user_agent,
                self.web_settings.GA_TRACKER_URL, self.web_settings.API_VERSION, host, path)
            # add the event, if applicable
            if event:
                request_body += '\nv=1&t=event&tid={}&ds=web&cid={}&uip={}&ua={}&an={}&av={}&dh={}&dp={}'.format(
                    self.web_settings.GA_ACCOUNT, this_user, remote_ip, user_agent,
                    self.web_settings.GA_TRACKER_URL, self.web_settings.API_VERSION, host, path)
                # add event information also
                request_body += '&ec={}&ea={}'.format(event['category'], event['action'])
                if event.get('label', False) and event.get('value', False):
                    request_body += '&el={}&ev={}'.format(event['label'], event['value'])

            req = HTTPRequest('http://www.google-analytics.com/batch', method='POST', body=request_body)

            #now send actual async requests
            http_client = AsyncHTTPClient()
            http_client.fetch(req)



"""
    Biothings Web Settings Default
"""


from .descriptions import KWARG_DESCRIPTIONS

# *****************************************************************************
# Elasticsearch Settings
# *****************************************************************************
# elasticsearch server transport url
ES_HOST = 'localhost:9200'
# load balancing by connecting to all nodes in the cluster
ES_SNIFF = False
# timeout for python es client (global request timeout)
ES_CLIENT_TIMEOUT = 120
# elasticsearch index name
ES_INDEX = '_all'
# elasticsearch document type for es<7, also biothing type
ES_DOC_TYPE = 'all'
# additional index support
ES_INDICES = {
    # "biothing_type_1": "index1",
    # "biothing_type_2": "index1,alias1,pattern_*"
}
# Amount of time a scroll request is kept open
ES_SCROLL_TIME = '1m'
# Size of each scroll request return
ES_SCROLL_SIZE = 1000
# Maximum size of result return
ES_SIZE_CAP = 1000
# Maximum result window => maximum for "from" parameter
ES_RESULT_WINDOW_SIZE_CAP = 10000

# *****************************************************************************
# Web Application & Base Handler
# *****************************************************************************
BIOTHING_TYPES = ()

# api version in the URL patterns and elsewhere
API_PREFIX = ''
API_VERSION = 'v1'

# project URL routing
APP_LIST = [
    (r"/", 'biothings.web.handlers.FrontPageHandler'),
    (r"/{pre}/status", 'biothings.web.handlers.StatusHandler'),
    (r"/{pre}/{ver}/spec/?", 'biothings.web.handlers.APISpecificationHandler'),
    (r"/{pre}/{ver}/{typ}/metadata/fields/?", 'biothings.web.handlers.MetadataFieldHandler'),
    (r"/{pre}/{ver}/{typ}/metadata/?", 'biothings.web.handlers.MetadataSourceHandler'),
    (r"/{pre}/{ver}/{typ}/query/?", 'biothings.web.handlers.QueryHandler'),
    (r"/{pre}/{ver}/{typ}/([^\/]+)/?", 'biothings.web.handlers.BiothingHandler'),
    (r"/{pre}/{ver}/{typ}/?", 'biothings.web.handlers.BiothingHandler'),
    (r"/{pre}/{ver}/metadata/fields/?", 'biothings.web.handlers.MetadataFieldHandler'),
    (r"/{pre}/{ver}/metadata/?", 'biothings.web.handlers.MetadataSourceHandler'),
    (r"/{pre}/{ver}/query/?", 'biothings.web.handlers.QueryHandler'),
]

# string used in headers to support CORS
ACCESS_CONTROL_ALLOW_METHODS = 'GET,POST,OPTIONS'
ACCESS_CONTROL_ALLOW_HEADERS = (
    'Content-Type, Depth, User-Agent, If-Modified-Since,'
    'Cache-Control, X-File-Size, X-Requested-With, X-File-Name'
)
# Caching behavior
DISABLE_CACHING = False
CACHE_MAX_AGE = 604800  # default 7 days

# Global default cap for list inputs
LIST_SIZE_CAP = 1000

# For format=html
HTML_OUT_HEADER_IMG = "https://biothings.io/static/favicon.ico"
HTML_OUT_TITLE = "<p>Biothings API</p>"
METADATA_DOCS_URL = "javascript:;"
QUERY_DOCS_URL = "javascript:;"
ANNOTATION_DOCS_URL = "javascript:;"

# path to the git repository for the app-specific code, override
APP_GIT_REPOSITORY = '.'

# default static path, relative to current working dir
# (from where app is launched)
STATIC_PATH = "static"

# color support is provided by tornado.log
LOGGING_FORMAT = "%(color)s[%(levelname)s %(name)s %(module)s:%(lineno)d]%(end_color)s %(message)s"

# *****************************************************************************
# User Input Control
# *****************************************************************************
COMMON_KWARGS = {
    # control group
    'raw': {'type': bool, 'default': False, 'group': 'control'},
    'rawquery': {'type': bool, 'default': False, 'group': 'control'},
    # esqb group
    '_source': {'type': list, 'group': 'esqb', 'max': 1000, 'alias': ['fields', 'field', 'filter']},
    'size': {'type': int, 'group': 'esqb', 'max': 1000, 'alias': 'limit'},
    # transform group
    'dotfield': {'type': bool, 'default': False, 'group': 'transform'},
    '_sorted': {'type': bool, 'default': True, 'group': 'transform'},  # alaphabetically
    'always_list': {'type': list, 'group': 'transform', 'max': 1000},
    'allow_null': {'type': list, 'group': 'transform', 'max': 1000}
}
ANNOTATION_KWARGS = {
    '*': COMMON_KWARGS.copy(),
    'GET': {'id': {'type': str, 'path': 0}},
    'POST': {'ids': {'type': list, 'max': 1000, 'required': True}}
}
QUERY_KWARGS = {
    '*': COMMON_KWARGS.copy(),
    'GET': {'q': {'type': str, 'default': '__all__', 'group': 'esqb'},
            'aggs': {'type': list, 'max': 1000, 'group': 'esqb', 'alias': 'facets'},
            'facet_size': {'type': int, 'default': 10, 'max': 1000, 'group': 'esqb', },
            'from': {'type': int, 'max': 10000, 'group': 'esqb', 'alias': 'skip'},
            'userquery': {'type': str, 'group': 'esqb', 'alias': ['userfilter']},
            'sort': {'type': list, 'group': 'esqb', 'max': 1000},
            'explain': {'type': bool, 'group': 'esqb'},
            'fetch_all': {'type': bool, 'group': 'es'},
            'scroll_id': {'type': str, 'group': 'es'}},
    'POST': {'q': {'type': list, 'required': True, 'group': 'esqb'},
             'scopes': {'type': list, 'default': ['_id'], 'group': 'esqb', 'max': 1000}}
}
# *****************************************************************************
# Elasticsearch Query Pipeline
# *****************************************************************************
ES_QUERY_BUILDER = 'biothings.web.pipeline.ESQueryBuilder'
# For the userquery folder for this app
USERQUERY_DIR = 'userquery'
# Allow the __any__ random doc retrieval
ALLOW_RANDOM_QUERY = False
# Allow facets to be nested with ( )
ALLOW_NESTED_AGGS = False

ES_QUERY_BACKEND = 'biothings.web.pipeline.ESQueryBackend'
ES_RESULT_TRANSFORM = 'biothings.web.pipeline.ESResultTransform'

# A list of fields to exclude from metadata/fields endpoint
AVAILABLE_FIELDS_EXCLUDED = ['all']
# A path to the available fields notes
AVAILABLE_FIELDS_NOTES_PATH = ''

LICENSE_TRANSFORM = {
    # "alias" :  "datasource",
    # "dot.field" :  "datasource"
}

# *****************************************************************************
# Analytics Settings
# *****************************************************************************

# Sentry project address
SENTRY_CLIENT_KEY = ''

# Google Analytics Account ID
GA_ACCOUNT = ''
GA_TRACKER_URL = 'mybiothing.info'  # TODO
GA_ACTION_QUERY_GET = 'query_get'
GA_ACTION_QUERY_POST = 'query_post'
GA_ACTION_ANNOTATION_GET = 'biothing_get'
GA_ACTION_ANNOTATION_POST = 'biothing_post'

# for standalone instance tracking
STANDALONE_TRACKING_URL = ''
# dictionary with AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY k,v for standalone user AWS IAM
STANDALONE_AWS_CREDENTIALS = {}
# batch size for standalone tracking (sending requests to AWS lambda)
STANDALONE_TRACKING_BATCH_SIZE = 1000


# *****************************************************************************
# Endpoints Specifics & Others
# *****************************************************************************
# Annotation #
ANNOTATION_DEFAULT_SCOPES = ['_id']
ANNOTATION_ID_REGEX_LIST = []  # [(re.compile(r'rs[0-9]+', re.I), 'dbsnp.rsid')]
#
# Status #
# https://www.elastic.co/guide/en/elasticsearch/reference/master/docs-get.html
STATUS_CHECK = {
    # 'id': '',
    # 'index': '',
    # 'doc_type': ''
}
#
# Biothing #
ID_REQUIRED_MESSAGE = 'ID required'
ID_NOT_FOUND_TEMPLATE = "ID '{bid}' not found"
